import { IQueryHandler, QueryHandler } from '@nestjs/cqrs';
import { MatchQuery, ScheduleItem } from './match.query';
import { Algorithm } from './algorithm.abstract';
import { PassengerOrientedAlgorithm } from './passenger-oriented-algorithm';
import { AlgorithmType } from '../../types/algorithm.types';
import { Inject } from '@nestjs/common';
import { AdRepositoryPort } from '@modules/ad/core/application/ports/ad.repository.port';
import {
  AD_CONFIGURATION_REPOSITORY,
  AD_REPOSITORY,
  INPUT_DATETIME_TRANSFORMER,
  MATCHING_REPOSITORY,
} from '@modules/ad/ad.di-tokens';
import { MatchEntity } from '@modules/ad/core/domain/match.entity';
import { DateTimeTransformerPort } from '../../ports/datetime-transformer.port';
import { Paginator } from '@mobicoop/ddd-library';
import { MatchingEntity } from '@modules/ad/core/domain/matching.entity';
import { MatchingRepositoryPort } from '../../ports/matching.repository.port';
import {
  Domain,
  Configurator,
  GetConfigurationRepositoryPort,
} from '@mobicoop/configuration-module';
import {
  CARPOOL_CONFIG_DEPARTURE_TIME_MARGIN,
  CARPOOL_CONFIG_ROLE,
  CARPOOL_CONFIG_SEATS_PROPOSED,
  CARPOOL_CONFIG_SEATS_REQUESTED,
  CARPOOL_CONFIG_STRICT_FREQUENCY,
  CarpoolKeyTypes,
} from '@modules/ad/ad.constants';
import {
  MATCH_CONFIG_ALGORITHM,
  MATCH_CONFIG_AZIMUTH_MARGIN,
  MATCH_CONFIG_MAX_DETOUR_DISTANCE_RATIO,
  MATCH_CONFIG_MAX_DETOUR_DURATION_RATIO,
  MATCH_CONFIG_PROPORTION,
  MATCH_CONFIG_REMOTENESS,
  MATCH_CONFIG_USE_AZIMUTH,
  MATCH_CONFIG_USE_PROPORTION,
  MatchConfig,
  PAGINATION_CONFIG_PER_PAGE,
  PaginationConfig,
} from '@modules/ad/match.constants';

@QueryHandler(MatchQuery)
export class MatchQueryHandler implements IQueryHandler {
  constructor(
    @Inject(AD_CONFIGURATION_REPOSITORY)
    private readonly configurationRepository: GetConfigurationRepositoryPort,
    @Inject(AD_REPOSITORY) private readonly adRepository: AdRepositoryPort,
    @Inject(MATCHING_REPOSITORY)
    private readonly matchingRepository: MatchingRepositoryPort,
    @Inject(INPUT_DATETIME_TRANSFORMER)
    private readonly datetimeTransformer: DateTimeTransformerPort,
  ) {}

  execute = async (query: MatchQuery): Promise<MatchingResult> => {
    const carpoolConfigurator: Configurator =
      await this.configurationRepository.mget(Domain.CARPOOL, CarpoolKeyTypes);
    const matchConfigurator: Configurator =
      await this.configurationRepository.mget(Domain.MATCH, MatchConfig);
    const paginationConfigurator: Configurator =
      await this.configurationRepository.mget(
        Domain.PAGINATION,
        PaginationConfig,
      );
    query
      .setMissingMarginDurations(
        carpoolConfigurator.get<number>(CARPOOL_CONFIG_DEPARTURE_TIME_MARGIN),
      )
      .setMissingStrict(
        carpoolConfigurator.get<boolean>(CARPOOL_CONFIG_STRICT_FREQUENCY),
      )
      .setDefaultDriverAndPassengerParameters({
        driver:
          carpoolConfigurator.get<string>(CARPOOL_CONFIG_ROLE) === 'driver',
        passenger:
          carpoolConfigurator.get<string>(CARPOOL_CONFIG_ROLE) === 'passenger',
        seatsProposed: carpoolConfigurator.get<number>(
          CARPOOL_CONFIG_SEATS_PROPOSED,
        ),
        seatsRequested: carpoolConfigurator.get<number>(
          CARPOOL_CONFIG_SEATS_REQUESTED,
        ),
      })
      .setDefaultAlgorithmParameters({
        algorithmType: matchConfigurator.get<AlgorithmType>(
          MATCH_CONFIG_ALGORITHM,
        ),
        remoteness: matchConfigurator.get<number>(MATCH_CONFIG_REMOTENESS),
        useProportion: matchConfigurator.get<boolean>(
          MATCH_CONFIG_USE_PROPORTION,
        ),
        proportion: matchConfigurator.get<number>(MATCH_CONFIG_PROPORTION),
        useAzimuth: matchConfigurator.get<boolean>(MATCH_CONFIG_USE_AZIMUTH),
        azimuthMargin: matchConfigurator.get<number>(
          MATCH_CONFIG_AZIMUTH_MARGIN,
        ),
        maxDetourDistanceRatio: matchConfigurator.get<number>(
          MATCH_CONFIG_MAX_DETOUR_DISTANCE_RATIO,
        ),
        maxDetourDurationRatio: matchConfigurator.get<number>(
          MATCH_CONFIG_MAX_DETOUR_DURATION_RATIO,
        ),
      })
      .setDefaultPagination({
        page: 1,
        perPage: paginationConfigurator.get<number>(PAGINATION_CONFIG_PER_PAGE),
      })
      .setDatesAndSchedule(this.datetimeTransformer);
    let matchingEntity: MatchingEntity | undefined = await this._cachedMatching(
      query.id,
    );
    if (!matchingEntity)
      matchingEntity = (await this._createMatching(query)) as MatchingEntity;
    const perPage: number = query.perPage as number;
    const page: number = Paginator.pageNumber(
      matchingEntity.getProps().matches.length,
      perPage,
      query.page as number,
    );
    return {
      id: matchingEntity.id,
      matches: Paginator.pageItems(
        matchingEntity.getProps().matches,
        page,
        perPage,
      ),
      total: matchingEntity.getProps().matches.length,
      page,
      perPage,
    };
  };

  private _cachedMatching = async (
    id?: string,
  ): Promise<MatchingEntity | undefined> => {
    if (!id) return undefined;
    try {
      return await this.matchingRepository.get(id);
    } catch (e: any) {
      return undefined;
    }
  };

  private _createMatching = async (
    query: MatchQuery,
  ): Promise<MatchingEntity> => {
    await query.setRoutes();

    let algorithm: Algorithm;
    switch (query.algorithmType) {
      case AlgorithmType.PASSENGER_ORIENTED:
      default:
        algorithm = new PassengerOrientedAlgorithm(query, this.adRepository);
    }

    const matches: MatchEntity[] = await algorithm.match();
    const matchingEntity = MatchingEntity.create({
      matches,
      query: {
        driver: query.driver as boolean,
        passenger: query.passenger as boolean,
        frequency: query.frequency,
        fromDate: query.fromDate,
        toDate: query.toDate,
        schedule: query.schedule?.map((scheduleItem: ScheduleItem) => ({
          day: scheduleItem.day as number,
          time: scheduleItem.time,
          margin: scheduleItem.margin as number,
        })),
        seatsProposed: query.seatsProposed as number,
        seatsRequested: query.seatsRequested as number,
        strict: query.strict as boolean,
        waypoints: query.waypoints,
        algorithmType: query.algorithmType as AlgorithmType,
        remoteness: query.remoteness as number,
        useProportion: query.useProportion as boolean,
        proportion: query.proportion as number,
        useAzimuth: query.useAzimuth as boolean,
        azimuthMargin: query.azimuthMargin as number,
        maxDetourDistanceRatio: query.maxDetourDistanceRatio as number,
        maxDetourDurationRatio: query.maxDetourDurationRatio as number,
      },
    });
    await this.matchingRepository.save(matchingEntity);
    return matchingEntity;
  };
}

export type MatchingResult = {
  id: string;
  matches: MatchEntity[];
  total: number;
  page: number;
  perPage: number;
};
