import { ArgumentInvalidException } from '@mobicoop/ddd-library';
import { Frequency, Role } from '@modules/ad/core/domain/ad.types';
import { CandidateEntity } from '@modules/ad/core/domain/candidate.entity';
import {
  SpacetimeDetourRatio,
  Target,
} from '@modules/ad/core/domain/candidate.types';
import { Actor } from '@modules/ad/core/domain/value-objects/actor.value-object';
import { CarpoolPathItemProps } from '@modules/ad/core/domain/value-objects/carpool-path-item.value-object';
import {
  Journey,
  JourneyProps,
} from '@modules/ad/core/domain/value-objects/journey.value-object';
import { PointProps } from '@modules/ad/core/domain/value-objects/point.value-object';
import { ScheduleItemProps } from '@modules/ad/core/domain/value-objects/schedule-item.value-object';
import { StepProps } from '@modules/ad/core/domain/value-objects/step.value-object';

const waypointsSet1: PointProps[] = [
  {
    lat: 48.678454,
    lon: 6.189745,
  },
  {
    lat: 48.84877,
    lon: 2.398457,
  },
];

const waypointsSet2: PointProps[] = [
  {
    lat: 48.689445,
    lon: 6.17651,
  },
  {
    lat: 48.8566,
    lon: 2.3522,
  },
];

const schedule1: ScheduleItemProps[] = [
  {
    day: 1,
    time: '07:00',
    margin: 900,
  },
];

const schedule2: ScheduleItemProps[] = [
  {
    day: 1,
    time: '07:10',
    margin: 900,
  },
];

const schedule3: ScheduleItemProps[] = [
  {
    day: 1,
    time: '06:30',
    margin: 900,
  },
  {
    day: 2,
    time: '06:30',
    margin: 900,
  },
  {
    day: 3,
    time: '06:00',
    margin: 900,
  },
  {
    day: 4,
    time: '06:30',
    margin: 900,
  },
  {
    day: 5,
    time: '06:30',
    margin: 900,
  },
];

const schedule4: ScheduleItemProps[] = [
  {
    day: 1,
    time: '06:50',
    margin: 900,
  },
  {
    day: 2,
    time: '06:50',
    margin: 900,
  },
  {
    day: 4,
    time: '06:50',
    margin: 900,
  },
  {
    day: 5,
    time: '06:50',
    margin: 900,
  },
];

const schedule5: ScheduleItemProps[] = [
  {
    day: 0,
    time: '00:02',
    margin: 900,
  },
  {
    day: 1,
    time: '07:05',
    margin: 900,
  },
];

const schedule6: ScheduleItemProps[] = [
  {
    day: 1,
    time: '23:10',
    margin: 900,
  },
  {
    day: 6,
    time: '23:57',
    margin: 900,
  },
];

const schedule7: ScheduleItemProps[] = [
  {
    day: 4,
    time: '19:00',
    margin: 900,
  },
];

const spacetimeDetourRatio: SpacetimeDetourRatio = {
  maxDistanceDetourRatio: 0.3,
  maxDurationDetourRatio: 0.3,
};

const carpoolPath1: CarpoolPathItemProps[] = [
  {
    lat: 48.689445,
    lon: 6.17651,
    actors: [
      new Actor({
        role: Role.DRIVER,
        target: Target.START,
      }),
      new Actor({
        role: Role.PASSENGER,
        target: Target.START,
      }),
    ],
  },
  {
    lat: 48.8566,
    lon: 2.3522,
    actors: [
      new Actor({
        role: Role.DRIVER,
        target: Target.FINISH,
      }),
      new Actor({
        role: Role.PASSENGER,
        target: Target.FINISH,
      }),
    ],
  },
];

const carpoolPath2: CarpoolPathItemProps[] = [
  {
    lat: 48.689445,
    lon: 6.17651,
    actors: [
      new Actor({
        role: Role.DRIVER,
        target: Target.START,
      }),
    ],
  },
  {
    lat: 48.678451,
    lon: 6.168784,
    actors: [
      new Actor({
        role: Role.DRIVER,
        target: Target.NEUTRAL,
      }),
      new Actor({
        role: Role.PASSENGER,
        target: Target.START,
      }),
    ],
  },
  {
    lat: 48.848715,
    lon: 2.36985,
    actors: [
      new Actor({
        role: Role.DRIVER,
        target: Target.NEUTRAL,
      }),
      new Actor({
        role: Role.PASSENGER,
        target: Target.FINISH,
      }),
    ],
  },
  {
    lat: 48.8566,
    lon: 2.3522,
    actors: [
      new Actor({
        role: Role.DRIVER,
        target: Target.FINISH,
      }),
    ],
  },
];

const steps: StepProps[] = [
  {
    lat: 48.689445,
    lon: 6.17651,
    duration: 0,
    distance: 0,
  },
  {
    lat: 48.678451,
    lon: 6.168784,
    duration: 1254,
    distance: 33462,
  },
  {
    lat: 48.848715,
    lon: 2.36985,
    duration: 12477,
    distance: 343654,
  },
  {
    lat: 48.8566,
    lon: 2.3522,
    duration: 13548,
    distance: 350145,
  },
];

describe('Candidate entity', () => {
  it('should create a new candidate entity', () => {
    const candidateEntity: CandidateEntity = CandidateEntity.create({
      id: 'cc260669-1c6d-441f-80a5-19cd59afb777',
      role: Role.DRIVER,
      frequency: Frequency.PUNCTUAL,
      dateInterval: {
        lowerDate: '2023-08-28',
        higherDate: '2023-08-28',
      },
      driverWaypoints: waypointsSet1,
      passengerWaypoints: waypointsSet2,
      driverDistance: 350145,
      driverDuration: 13548,
      driverSchedule: schedule1,
      passengerSchedule: schedule2,
      spacetimeDetourRatio,
    });
    expect(candidateEntity.id.length).toBe(36);
  });

  it('should set a candidate entity carpool path', () => {
    const candidateEntity: CandidateEntity = CandidateEntity.create({
      id: '5600ccfb-ab69-4d03-aa30-0fbe84fcedc0',
      role: Role.PASSENGER,
      frequency: Frequency.PUNCTUAL,
      dateInterval: {
        lowerDate: '2023-08-28',
        higherDate: '2023-08-28',
      },
      driverWaypoints: waypointsSet2,
      passengerWaypoints: waypointsSet2,
      driverDistance: 350145,
      driverDuration: 13548,
      driverSchedule: schedule1,
      passengerSchedule: schedule2,
      spacetimeDetourRatio,
    }).setCarpoolPath(carpoolPath1);
    expect(candidateEntity.getProps().carpoolPath).toHaveLength(2);
  });

  it('should create a new candidate entity with spacetime metrics', () => {
    const candidateEntity: CandidateEntity = CandidateEntity.create({
      id: 'cc260669-1c6d-441f-80a5-19cd59afb777',
      role: Role.DRIVER,
      frequency: Frequency.PUNCTUAL,
      dateInterval: {
        lowerDate: '2023-08-28',
        higherDate: '2023-08-28',
      },
      driverWaypoints: waypointsSet1,
      passengerWaypoints: waypointsSet2,
      driverDistance: 350145,
      driverDuration: 13548,
      driverSchedule: schedule1,
      passengerSchedule: schedule2,
      spacetimeDetourRatio,
    }).setMetrics(352688, 14587);
    expect(candidateEntity.getProps().distance).toBe(352688);
    expect(candidateEntity.getProps().duration).toBe(14587);
  });

  describe('detour validation', () => {
    it('should not validate a candidate entity with exceeding distance detour', () => {
      const candidateEntity: CandidateEntity = CandidateEntity.create({
        id: 'cc260669-1c6d-441f-80a5-19cd59afb777',
        role: Role.DRIVER,
        frequency: Frequency.PUNCTUAL,
        dateInterval: {
          lowerDate: '2023-08-28',
          higherDate: '2023-08-28',
        },
        driverWaypoints: waypointsSet1,
        passengerWaypoints: waypointsSet2,
        driverDistance: 350145,
        driverDuration: 13548,
        driverSchedule: schedule1,
        passengerSchedule: schedule2,
        spacetimeDetourRatio,
      }).setMetrics(458690, 13980);
      expect(candidateEntity.isDetourValid()).toBeFalsy();
    });
    it('should not validate a candidate entity with exceeding duration detour', () => {
      const candidateEntity: CandidateEntity = CandidateEntity.create({
        id: 'cc260669-1c6d-441f-80a5-19cd59afb777',
        role: Role.DRIVER,
        frequency: Frequency.PUNCTUAL,
        dateInterval: {
          lowerDate: '2023-08-28',
          higherDate: '2023-08-28',
        },
        driverWaypoints: waypointsSet1,
        passengerWaypoints: waypointsSet2,
        driverDistance: 350145,
        driverDuration: 13548,
        driverSchedule: schedule1,
        passengerSchedule: schedule2,
        spacetimeDetourRatio,
      }).setMetrics(352368, 18314);
      expect(candidateEntity.isDetourValid()).toBeFalsy();
    });
  });

  describe('Journeys', () => {
    it('should create journeys for a single date', () => {
      const candidateEntity: CandidateEntity = CandidateEntity.create({
        id: '5600ccfb-ab69-4d03-aa30-0fbe84fcedc0',
        role: Role.PASSENGER,
        frequency: Frequency.PUNCTUAL,
        dateInterval: {
          lowerDate: '2023-08-28',
          higherDate: '2023-08-28',
        },
        driverWaypoints: waypointsSet1,
        passengerWaypoints: waypointsSet2,
        driverDistance: 350145,
        driverDuration: 13548,
        driverSchedule: schedule1,
        passengerSchedule: schedule2,
        spacetimeDetourRatio,
      })
        .setCarpoolPath(carpoolPath2)
        .setSteps(steps)
        .createJourneys();
      expect(candidateEntity.getProps().journeys).toHaveLength(1);
    });
    it('should create journeys for a single date without driver schedule', () => {
      const candidateEntity: CandidateEntity = CandidateEntity.create({
        id: '5600ccfb-ab69-4d03-aa30-0fbe84fcedc0',
        role: Role.PASSENGER,
        frequency: Frequency.PUNCTUAL,
        dateInterval: {
          lowerDate: '2023-08-28',
          higherDate: '2023-08-28',
        },
        driverWaypoints: waypointsSet1,
        passengerWaypoints: waypointsSet2,
        driverDistance: 350145,
        driverDuration: 13548,
        driverSchedule: undefined,
        passengerSchedule: schedule2,
        spacetimeDetourRatio,
      })
        .setCarpoolPath(carpoolPath2)
        .setSteps(steps)
        .createJourneys();
      expect(candidateEntity.getProps().journeys).toHaveLength(1);
      // computed driver start time should be 06:49
      expect(
        (
          candidateEntity.getProps().journeys as JourneyProps[]
        )[0].journeyItems[0].actorTimes[0].firstDatetime.getUTCMinutes(),
      ).toBe(49);
      expect(
        (
          candidateEntity.getProps().journeys as JourneyProps[]
        )[0].journeyItems[0].actorTimes[0].firstDatetime.getUTCHours(),
      ).toBe(6);
    });
    it('should create journeys for a single date without passenger schedule', () => {
      const candidateEntity: CandidateEntity = CandidateEntity.create({
        id: '5600ccfb-ab69-4d03-aa30-0fbe84fcedc0',
        role: Role.PASSENGER,
        frequency: Frequency.PUNCTUAL,
        dateInterval: {
          lowerDate: '2023-08-28',
          higherDate: '2023-08-28',
        },
        driverWaypoints: waypointsSet1,
        passengerWaypoints: waypointsSet2,
        driverDistance: 350145,
        driverDuration: 13548,
        driverSchedule: schedule1,
        passengerSchedule: undefined,
        spacetimeDetourRatio,
      })
        .setCarpoolPath(carpoolPath2)
        .setSteps(steps)
        .createJourneys();
      expect(candidateEntity.getProps().journeys).toHaveLength(1);
      // computed passenger start time should be 07:20
      expect(
        (
          candidateEntity.getProps().journeys as JourneyProps[]
        )[0].journeyItems[1].actorTimes[0].firstDatetime.getUTCMinutes(),
      ).toBe(20);
      expect(
        (
          candidateEntity.getProps().journeys as JourneyProps[]
        )[0].journeyItems[1].actorTimes[0].firstDatetime.getUTCHours(),
      ).toBe(7);
    });
    it('should throw without driver and passenger schedule', () => {
      expect(() =>
        CandidateEntity.create({
          id: '5600ccfb-ab69-4d03-aa30-0fbe84fcedc0',
          role: Role.PASSENGER,
          frequency: Frequency.PUNCTUAL,
          dateInterval: {
            lowerDate: '2023-08-28',
            higherDate: '2023-08-28',
          },
          driverWaypoints: waypointsSet1,
          passengerWaypoints: waypointsSet2,
          driverDistance: 350145,
          driverDuration: 13548,
          driverSchedule: undefined,
          passengerSchedule: undefined,
          spacetimeDetourRatio,
        })
          .setCarpoolPath(carpoolPath2)
          .setSteps(steps)
          .createJourneys(),
      ).toThrow(ArgumentInvalidException);
    });
    it('should create journeys for multiple dates', () => {
      const candidateEntity: CandidateEntity = CandidateEntity.create({
        id: '5600ccfb-ab69-4d03-aa30-0fbe84fcedc0',
        role: Role.PASSENGER,
        frequency: Frequency.RECURRENT,
        dateInterval: {
          lowerDate: '2023-09-01',
          higherDate: '2024-09-01',
        },
        driverWaypoints: waypointsSet1,
        passengerWaypoints: waypointsSet2,
        driverDistance: 350145,
        driverDuration: 13548,
        driverSchedule: schedule3,
        passengerSchedule: schedule4,
        spacetimeDetourRatio,
      })
        .setCarpoolPath(carpoolPath2)
        .setSteps(steps)
        .createJourneys();
      expect(candidateEntity.getProps().journeys).toHaveLength(4);
      expect(
        (
          candidateEntity.getProps().journeys as Journey[]
        )[0].firstDate.getDate(),
      ).toBe(4);
      expect(
        (
          candidateEntity.getProps().journeys as Journey[]
        )[0].journeyItems[1].actorTimes[1].firstMinDatetime.getDate(),
      ).toBe(4);
      expect(
        (
          candidateEntity.getProps().journeys as Journey[]
        )[1].journeyItems[1].actorTimes[1].firstMinDatetime.getDate(),
      ).toBe(5);
    });
    it('should create journeys for multiple dates, including week edges (saturday/sunday)', () => {
      const candidateEntity: CandidateEntity = CandidateEntity.create({
        id: '5600ccfb-ab69-4d03-aa30-0fbe84fcedc0',
        role: Role.PASSENGER,
        frequency: Frequency.RECURRENT,
        dateInterval: {
          lowerDate: '2023-09-01',
          higherDate: '2024-09-01',
        },
        driverWaypoints: waypointsSet1,
        passengerWaypoints: waypointsSet2,
        driverDistance: 350145,
        driverDuration: 13548,
        driverSchedule: schedule5,
        passengerSchedule: schedule6,
        spacetimeDetourRatio,
      })
        .setCarpoolPath(carpoolPath2)
        .setSteps(steps)
        .createJourneys();
      expect(candidateEntity.getProps().journeys).toHaveLength(1);
      expect(
        (candidateEntity.getProps().journeys as Journey[])[0].journeyItems[1]
          .actorTimes[0].target,
      ).toBe(Target.NEUTRAL);
      expect(
        (
          candidateEntity.getProps().journeys as Journey[]
        )[0].journeyItems[1].actorTimes[0].firstDatetime.getUTCHours(),
      ).toBe(0);
      expect(
        (
          candidateEntity.getProps().journeys as Journey[]
        )[0].journeyItems[1].actorTimes[0].firstDatetime.getUTCMinutes(),
      ).toBe(22);
      expect(
        (
          candidateEntity.getProps().journeys as Journey[]
        )[0].journeyItems[1].actorTimes[1].firstMinDatetime.getUTCHours(),
      ).toBe(23);
      expect(
        (
          candidateEntity.getProps().journeys as Journey[]
        )[0].journeyItems[1].actorTimes[1].firstMinDatetime.getUTCMinutes(),
      ).toBe(42);
    });

    it('should not create journeys if dates does not match', () => {
      const candidateEntity: CandidateEntity = CandidateEntity.create({
        id: '5600ccfb-ab69-4d03-aa30-0fbe84fcedc0',
        role: Role.PASSENGER,
        frequency: Frequency.RECURRENT,
        dateInterval: {
          lowerDate: '2023-09-01',
          higherDate: '2024-09-01',
        },
        driverWaypoints: waypointsSet1,
        passengerWaypoints: waypointsSet2,
        driverDistance: 350145,
        driverDuration: 13548,
        driverSchedule: schedule1,
        passengerSchedule: schedule7,
        spacetimeDetourRatio,
      })
        .setCarpoolPath(carpoolPath2)
        .setSteps(steps)
        .createJourneys();
      expect(candidateEntity.getProps().journeys).toHaveLength(0);
      expect(candidateEntity.hasJourneys()).toBeFalsy();
    });

    it('should not verify journeys if journeys is undefined', () => {
      const candidateEntity: CandidateEntity = CandidateEntity.create({
        id: '5600ccfb-ab69-4d03-aa30-0fbe84fcedc0',
        role: Role.PASSENGER,
        frequency: Frequency.RECURRENT,
        dateInterval: {
          lowerDate: '2023-09-01',
          higherDate: '2024-09-01',
        },
        driverWaypoints: waypointsSet1,
        passengerWaypoints: waypointsSet2,
        driverDistance: 350145,
        driverDuration: 13548,
        driverSchedule: schedule1,
        passengerSchedule: schedule7,
        spacetimeDetourRatio,
      })
        .setCarpoolPath(carpoolPath2)
        .setSteps(steps);
      expect(candidateEntity.hasJourneys()).toBeFalsy();
    });
  });
});
